// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.logic.ldt;

import de.uka.ilkd.key.java.Expression;
import de.uka.ilkd.key.java.abstraction.Type;
import de.uka.ilkd.key.java.expression.Literal;
import de.uka.ilkd.key.java.expression.Operator;
import de.uka.ilkd.key.java.expression.literal.CharLiteral;
import de.uka.ilkd.key.java.expression.literal.IntLiteral;
import de.uka.ilkd.key.java.expression.literal.LongLiteral;
import de.uka.ilkd.key.java.expression.operator.*;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.logic.op.TermSymbol;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.ExtList;


/** This class inherits from LDT and implements all method that are
 * necessary to handle integers, shorts and bytes.
 * It loads the needed rule file and offers method to convert java
 * number types to their logic counterpart
 */
public abstract class AbstractIntegerLDT extends LDT {

    /**
     * Public name constants
     */
    public static final String NEGATIVE_LITERAL_STRING = "neglit";
    public static final Name NUMBERS_NAME = new Name("Z");
    public static final Name CHAR_ID_NAME = new Name("C"); 
        
    /** the function symbols used to represent literals */
    protected final Function numbers;
    protected final Function charID;
    protected final Function sharp;
    protected final Function numberSymbol[] = new Function[10];

    /** the function symbols for the integer operations. 
     * Used as shortcuts.
     */
    protected final Function greaterThan;    
    protected final Function greaterOrEquals;
    protected final Function lessThan;
    protected final Function lessOrEquals;
    protected final Function plus;
    protected final Function minus;
    protected final Function times;
    protected final Function divide;
    protected final Function modulo;
    protected final Function jDivide;
    protected final Function jModulo;
    protected final Function negative;
    protected final Function negativeNumber;    
    protected final Function unaryMinusJint;
    protected final Function unaryMinusJlong;
    protected final Function addJint;
    protected final Function addJlong;
    protected final Function subJint;
    protected final Function subJlong;
    protected final Function mulJint;
    protected final Function mulJlong;
    protected final Function modJint;
    protected final Function modJlong;
    protected final Function divJint;
    protected final Function divJlong;
    protected final Function shiftrightJint;
    protected final Function shiftrightJlong;
    protected final Function shiftleftJint;
    protected final Function shiftleftJlong;
    protected final Function unsignedshiftrightJint;
    protected final Function unsignedshiftrightJlong;
    protected final Function orJint;
    protected final Function orJlong;
    protected final Function andJint;
    protected final Function andJlong;
    protected final Function xorJint;
    protected final Function xorJlong;
    protected final Function negJint;
    protected final Function negJlong;
   
    /** the predicate symbols for being within boundaries*/
    protected final Function inByte;
    protected Function inShort;
    protected Function inInt;
    protected Function inLong;
    protected Function inChar;

    /**
     * TODO: These fields are used, but never initialised. 
     */
    protected final Function shiftLeft = null; //TODO
    protected final Function shiftRight = null; //TODO
    protected final Function binaryAnd = null; //TODO
    protected final Function binaryNot = null; //TODO
    protected final Function binaryOr = null; //TODO
    protected final Function binaryXOr = null; //TODO
    


    protected AbstractIntegerLDT(Name name, 
	    			 Namespace sorts, 
	    			 Namespace functions, 
	    			 Type javaType) {
	super(name, sorts, javaType);
	
	numbers = addFunction((Function)functions.lookup(NUMBERS_NAME));	
        charID = addFunction((Function) functions.lookup(CHAR_ID_NAME));
        sharp   = addFunction((Function)functions.lookup(new Name("#")));
        assert sharp.sort() == numbers.argSort(0);
        for (int i = 0; i < 10; i++) {
            numberSymbol[i] = addFunction((Function)functions.lookup(new Name(""+i)));
            assert numberSymbol[i].sort() == numbers.argSort(0);
        }        
        
        greaterThan     = addFunction((Function)functions.lookup(new Name("gt")));
        greaterOrEquals = addFunction((Function)functions.lookup(new Name("geq")));
        lessThan        = addFunction((Function)functions.lookup(new Name("lt")));
        lessOrEquals    = addFunction((Function)functions.lookup(new Name("leq")));
        plus            = addFunction((Function)functions.lookup(new Name("add")));
        minus           = addFunction((Function)functions.lookup(new Name("sub")));
        times           = addFunction((Function)functions.lookup(new Name("mul")));
        divide          = addFunction((Function)functions.lookup(new Name("div")));
        modulo          = addFunction((Function)functions.lookup(new Name("mod")));
        jDivide         = addFunction((Function)functions.lookup(new Name("div")));
        jModulo         = addFunction((Function)functions.lookup(new Name("mod")));                  
        negative        = addFunction((Function)functions.lookup(new Name("neg")));
        negativeNumber  = addFunction((Function)functions.lookup(new Name(NEGATIVE_LITERAL_STRING)));
        unaryMinusJint  = addFunction((Function)functions.lookup(new Name("unaryMinusJint")));
        unaryMinusJlong = addFunction((Function)functions.lookup(new Name("unaryMinusJlong")));
        addJint         = addFunction((Function)functions.lookup(new Name("addJint")));
        addJlong        = addFunction((Function)functions.lookup(new Name("addJlong")));
        subJint         = addFunction((Function)functions.lookup(new Name("subJint")));
        subJlong        = addFunction((Function)functions.lookup(new Name("subJlong")));
        mulJint         = addFunction((Function)functions.lookup(new Name("mulJint")));
        mulJlong        = addFunction((Function)functions.lookup(new Name("mulJlong")));
        modJint         = addFunction((Function)functions.lookup(new Name("modJint")));
        modJlong        = addFunction((Function)functions.lookup(new Name("modJlong")));
        divJint         = addFunction((Function)functions.lookup(new Name("divJint")));
        divJlong        = addFunction((Function)functions.lookup(new Name("divJlong")));
        shiftrightJint  = addFunction((Function)functions.lookup(new Name("shiftrightJint")));
        shiftrightJlong = addFunction((Function)functions.lookup(new Name("shiftrightJlong")));
        shiftleftJint   = addFunction((Function)functions.lookup(new Name("shiftleftJint")));
        shiftleftJlong  = addFunction((Function)functions.lookup(new Name("shiftleftJlong")));
        unsignedshiftrightJint  =
            addFunction((Function)functions.lookup(new Name("unsignedshiftrightJint")));
        unsignedshiftrightJlong =
            addFunction((Function)functions.lookup(new Name("unsignedshiftrightJlong")));
        orJint          = addFunction((Function)functions.lookup(new Name("orJint")));
        orJlong         = addFunction((Function)functions.lookup(new Name("orJlong")));
        andJint         = addFunction((Function)functions.lookup(new Name("andJint")));
        andJlong        = addFunction((Function)functions.lookup(new Name("andJlong")));
        xorJint         = addFunction((Function)functions.lookup(new Name("xorJint")));
        xorJlong        = addFunction((Function)functions.lookup(new Name("xorJlong")));
        negJint         = addFunction((Function)functions.lookup(new Name("negJint")));
        negJlong        = addFunction((Function)functions.lookup(new Name("negJlong")));          

        inByte  = addFunction((Function)functions.lookup(new Name("inByte")));
        inShort = addFunction((Function)functions.lookup(new Name("inShort")));
        inInt   = addFunction((Function)functions.lookup(new Name("inInt")));
        inLong  = addFunction((Function)functions.lookup(new Name("inLong")));
        inChar  = addFunction((Function)functions.lookup(new Name("inChar")));
    }
        
    
    /** returns the function symbol for the given operation 
     * null if no function is found for the given operator
     * @return  the function symbol for the given operation 
    */
    public Function getFunctionFor
	(de.uka.ilkd.key.java.expression.Operator op) {
	if (op instanceof GreaterThan) {
	    return greaterThan;
	} else if (op instanceof GreaterOrEquals) {
	    return greaterOrEquals;
	} else if (op instanceof LessThan) {
	    return lessThan;
	} else if (op instanceof LessOrEquals) {
	    return lessOrEquals;
	} else if (op instanceof Divide) {
	    return jDivide;
	} else if (op instanceof Times) {
	    return times;
	} else if (op instanceof Plus) {
	    return plus;
	} else if (op instanceof Minus) {
	    return minus;
	} else if (op instanceof Modulo) {
	    return jModulo;
	} else if (op instanceof ShiftLeft) {
	    return shiftLeft;
	} else if (op instanceof ShiftRight) {
	    return shiftRight;
	} else if (op instanceof BinaryAnd) {
	    return binaryAnd;
	} else if (op instanceof BinaryNot) {
	    return binaryNot;
	} else if (op instanceof BinaryOr) {
	    return binaryOr;
	} else if (op instanceof BinaryXOr) {
	    return binaryXOr;
	} else if (op instanceof Negative) {
	    return negative;
	}
	return null;
    }
    
    
    /** returns true if the LDT offers an operation for the given java
     * operator and the logic subterms 
     * @param op the de.uka.ilkd.key.java.expression.Operator to
     * translate
     * @param subs the logic subterms of the java operator
     * @return  true if the LDT offers an operation for the given java
     * operator and the subterms 
     */
    public boolean isResponsible(Operator op, Term[] subs) {
        if (subs.length == 1) {
            return isResponsible(op, subs[0]);
        } else if (subs.length == 2) {
            return isResponsible(op, subs[0], subs[1]); 
        }
        return false;   
    }
    

    /** returns true if the LDT offers an operation for the given
     * binary java operator and the logic subterms 
     * @param op the de.uka.ilkd.key.java.expression.Operator to
     * translate
     * @param left the left subterm of the java operator
     * @param right the right subterm of the java operator
     * @return  true if the LDT offers an operation for the given java
     * operator and the subterms 
     */
    public boolean isResponsible(Operator op, Term left, Term right) {
        if (left!=null && left.sort().extendsTrans(targetSort()) 
                && right!=null && right.sort().extendsTrans(targetSort())) {
            if (getFunctionFor(op) != null) {
                return true;
            }
        }
        return false;
    }
    
    
    /** returns true if the LDT offers an operation for the given
     * unary java operator and the logic subterms 
     * @param op the de.uka.ilkd.key.java.expression.Operator to
     * translate
     * @param sub the logic subterms of the java operator
     * @return  true if the LDT offers an operation for the given java
     * operator and the subterm
     */
    public boolean isResponsible(Operator op, Term sub) {
        if (sub != null && sub.sort().extendsTrans(targetSort())) {
            if (op instanceof Negative) {
                return true;
            }
        }
        return false;
    }
    

    /** translates a given integer literal to its logic counterpart 
     * @param lit the Literal to be translated (has to be an
     * IntLiteral of an LongLiteral
     * @result the Term that represent the given integer in its logic
     * form
     */ 
    public Term translateLiteral(Literal lit) {
	int length=0;
	boolean minusFlag = false;
	Debug.assertTrue(lit instanceof IntLiteral || 
			 lit instanceof LongLiteral ||
			 lit instanceof CharLiteral,
			 "Literal '"+lit+"' is not an integer literal.");

	char[] int_ch=null;
	assert sharp != null;
	Term result = TermFactory.DEFAULT.createFunctionTerm(sharp);

	Function identifier=numbers;
	if (lit instanceof CharLiteral) {
	    lit = new IntLiteral(""+ (int)(((CharLiteral)lit)
					   .getCharValue()) ) ;
	    identifier = charID;
	}

	String literalString = ""; 
       	if (lit instanceof IntLiteral) {
	    literalString = ((IntLiteral)lit).getValue();
	} else {
	    Debug.assertTrue(lit instanceof LongLiteral);
	    literalString = ((LongLiteral)lit).getValue();
	}

	if (literalString.charAt(0) == '-') {
	    minusFlag = true;	    
	    literalString = 
		literalString.substring(1);
	}
       	if (lit instanceof IntLiteral) {
	    if (literalString.startsWith("0x")) {
                try {
                    int i = Integer.parseInt
			(literalString.substring(2),16);
                    int_ch=(""+i).toCharArray();
                }catch(NumberFormatException nfe) {
                    Debug.fail("Not a hexadecimal constant!");
		}
            } else {
	        int_ch = literalString.toCharArray();
            }            
            length = int_ch.length; 
	} else if (lit instanceof LongLiteral) {
	    if (literalString.startsWith("0x")) {
		try {
		    // long literals have an 'L' as last sign; we have
		    // to skip it 
                    final long l = Long.parseLong
			(literalString.substring(2, 
						 literalString.length() - 1), 
			 16);
                    int_ch=(""+l).toCharArray();
                } catch (NumberFormatException nfe) {
                    Debug.fail("Not a hexadecimal constant!");
		}
		length = int_ch.length; 
	    } else {
		// long literals have an 'L' as last sign; skip it
		int_ch = literalString.toCharArray();
		length = int_ch.length - 1; 
	    }
	}
	
        for (int i = 0; i < length; i++) {
	    result = TermFactory.DEFAULT.createFunctionTerm
		(numberSymbol[int_ch[i]-'0'], result);
        }
	if (minusFlag) {
	    result = TermFactory.DEFAULT.createFunctionTerm
		(negativeNumber, result);
	}
 	result = TermFactory.DEFAULT.createFunctionTerm
 	    (identifier, result);

	Debug.out("integerldt: result of translating literal (lit, result):", 
		  lit, result);

	return result;

    }

    
    public Function getNumberTerminator(){
        return sharp;
    }
    
    
    public Function getNumberSymbol() {
	return numbers;
    }
    
    
    public Function getCharSymbol() {
	return charID;
    }

    
    public Function getNumberLiteralFor(int number) {
        if (number < 0 || number > 9) {
            throw new IllegalArgumentException
            ("Number literal symbols range from 0 to 9. Requested was:" + number);
        }
        
        return numberSymbol[number];
    }
    
    private boolean isNumberLiteral(Function f) {
	char c=f.name().toString().charAt(0);
	return (c-'0'>=0) && (c-'0'<=9);
    }

    public boolean hasLiteralFunction(Function f) {
	return containsFunction(f) && (f.arity()==0 || isNumberLiteral(f));
    }

    public Expression translateTerm(Term t, ExtList children) {
	if (!containsFunction((Function) t.op())) return null;
	Function f = (Function)t.op();	
	if (isNumberLiteral(f) || f == numbers || f ==charID) {	    
	    StringBuffer sb = new StringBuffer("");
	    Term it = t;
	    if (f == charID || f == numbers) {
		it = it.sub(0); 
		f = (Function)it.op();	    
	    }
	    while (isNumberLiteral(f)) {
		sb.insert(0, f.name().toString().charAt(0));
		it=it.sub(0);
		f = (Function)it.op();	    
	    }
	    // numbers must end with a sharp
	    if (f == sharp) {
		return new IntLiteral(sb.toString());
	    }
	}
	throw new RuntimeException("IntegerLDT: Cannot convert term to program: "
				   +t);
    }
    
    /**
     * returns the unary function symbol for representing a negative
     * number literal
     * @return the function symbol used to represent negative number literals
     */
    public Function getNegativeNumberSign() {
        return negativeNumber;
    }

    
    /**
     * returns the unary function symbol for representing the negative value
     * e.g. <code> -(var) </code> 
     * @return the function symbol used to represent the negative sign
     */
    public Function getNegativeSign() {
        return negative;
    }
        
    /**
     * returns the function symbol used to represent addition of 
     * arithmetical integers
     * @return the function symbol for integer addition
     */
    public Function getArithAddition() {
        return plus;
    }

    /**
     * returns the function symbol used to represent substraction of 
     * arithmetical integers
     * @return the function symbol for integer substraction
     */
    public Function getArithSubstraction() {
        return minus;
    }

    
    /**
     * returns the function symbol used to represent multiplication on
     * the arithmetical integers
     * @return the function symbol used to represent integer multiplication
     */
    public Function getArithMultiplication() {
        return times;
    }
    
    /**
     * returns the function symbol used to represent division of
     * the arithmetical integers
     * @return the function symbol used to represent integer division
     */
    public Function getArithDivision() {
        return divide;
    }
    
    /**
     * returns the function symbol used to represent the modulo operation of
     * the arithmetical integers
     * @return the function symbol used to represent the integer modulo operation 
     */
    public TermSymbol getArithModulo() {        
        return modulo;
    }

    
    /**
     * returns the boolean function symbol to compare two integer values
     * <code>val1, val2</code> if <code>val1</code> is lesser or equal than
     * <code>val2</val> 
     * @return the boolean function symbol to compare two integer values
     */
    public Function getLessOrEquals() {
        return lessOrEquals;
    }

    /**
     * returns the boolean function symbol to compare two integer values
     * <code>val1, val2</code> if <code>val1</code> is lesser than
     * <code>val2</val> 
     * @return the boolean function symbol to compare two integer values
     */
    public Function getLessThan() {
        return lessThan;
    }

    /**
     * returns the boolean function symbol to compare two integer values
     * <code>val1, val2</code> if <code>val1</code> is greater or equals
     * <code>val2</val> 
     * @return the boolean function symbol to compare two integer values
     */
    public Function getGreaterOrEquals() {
        return greaterOrEquals;
    }

    /**
     * returns the boolean function symbol to compare two integer values
     * <code>val1, val2</code> if <code>val1</code> is greater than
     * <code>val2</val> 
     * @return the boolean function symbol to compare two integer values
     */
    public Function getGreaterThan() {
        return greaterThan;
    }
    
    /** returns a function mapping an arithmetic integer to its Java long representation */ 
    public Function getModuloLong() {       
        return modJlong;
    }
    
    public Function getJavaIntegerAddition() {        
        return addJint;
    }
    
    public abstract Function getInBoundsPredicate();
} 
