// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.strategy;

import de.uka.ilkd.key.logic.FormulaChangeInfo;
import de.uka.ilkd.key.logic.ListOfFormulaChangeInfo;
import de.uka.ilkd.key.logic.PIOPathIterator;
import de.uka.ilkd.key.logic.PosInOccurrence;
import de.uka.ilkd.key.logic.op.IUpdateOperator;
import de.uka.ilkd.key.logic.op.Operator;
import de.uka.ilkd.key.proof.FormulaTag;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.rule.RuleApp;
import de.uka.ilkd.key.util.Debug;

/**
 * Instances of this class are immutable
 */
public class FindTacletAppContainer extends TacletAppContainer {

    /**
     * The position of the rule app in two different representations:
     * <code>positionTag</code> denotes the concerned formula and survives
     * modifications of the sequent and of parts of the formula, and
     * <code>applicationPosition</code> is the original position for which the
     * rule app was created
     */
    private final FormulaTag      positionTag;
    private final PosInOccurrence applicationPosition;

    FindTacletAppContainer ( RuleApp         p_app,
			     PosInOccurrence p_pio,
			     RuleAppCost     p_cost,
			     Goal            p_goal,
                             long            p_age ) {
	super ( p_app, p_cost, p_age );
    	applicationPosition = p_pio;
    	positionTag         =
    	    p_goal.getFormulaTagManager().getTagForPos(p_pio.topLevel());

        if ( positionTag == null )
            // faster than <code>assertFalse</code>
            Debug.fail ( "Formula " + p_pio + " does not exist" );
    }


    /**
     * @return true iff the stored rule app is applicable for the given sequent,
     * i.e. if the find-position does still exist (if-formulas are not
     * considered)
     */
    protected boolean isStillApplicable ( Goal p_goal ) {
    	final PosInOccurrence topPos =
    	    p_goal.getFormulaTagManager().getPosForTag(positionTag);
	if ( topPos == null )
	    // the formula does not exist anymore, bail out
	    return false;
	if ( !topPos.constrainedFormula().constraint().equals
	     ( applicationPosition.constrainedFormula().constraint() ) )
	    // the constraint of the formula has changed, bail out
	    return false;
	if ( subformulaOrPreceedingUpdateHasChanged ( p_goal ) )
	    return false;
	if (!p_goal.proof().mgt().ruleApplicable(getRuleApp(), p_goal)) {
		return false;
	}
	return true;
    }


    /**
     * @return true iff a subformula that contains the find position stored by
     * this object has been altered since the creation of this object or if a 
     * preceeding update has changed
     */
    private boolean subformulaOrPreceedingUpdateHasChanged ( Goal p_goal ) {
    	ListOfFormulaChangeInfo infoList =
    	    p_goal.getFormulaTagManager().getModifications(positionTag);

	while ( !infoList.isEmpty () ) {
	    final FormulaChangeInfo info = infoList.head ();
	    infoList = infoList.tail ();
	    
	    if ( info.getNewFormula() ==
		 applicationPosition.constrainedFormula() )
		// then there were no relevant modifications since the creation
		// of the rule app object
		return false;

	    if ( !independantSubformulas ( info.getPositionOfModification() ) )
	        return true;	    
	}
	
	return false;
    }


    /**
     * checks if the modification path and the postition where this taclet application
     * has been matched again denote independant subformulas. The modification affects 
     * a formula <code>F</code> if <code>F</code> is a subformula of the modified one 
     * or the modification took part inside an update which may occur in the update 
     * prefix instantiation of the taclet application    
     * @return true iff <code>applicationPosition</code> is in the scope of
     * the position <code>p_pos</code> (the formulas are not compared, only the
     * positions within the formulas) and no indirect relationship exists which 
     * is established by a modification that occurred inside an update 
     */
    private boolean independantSubformulas ( PosInOccurrence p_pos ) {
    	final PIOPathIterator it0 = p_pos.iterator();
	final PIOPathIterator it1 = applicationPosition.iterator();
	
	while ( true ) {
	    final int index0 = it0.next ();
	    final int index1 = it1.next ();
            
            final Operator modOp = it0.getSubTerm().op();
            
            if (modOp instanceof IUpdateOperator) {                
                final int modTargetPos = ((IUpdateOperator)modOp).targetPos();                
                final Operator origOp  = it1.getSubTerm().op();                
                if (!(origOp instanceof IUpdateOperator) ||
                        (it0.getChild() < modTargetPos && 
                                it1.getChild() == ((IUpdateOperator)origOp).targetPos())) {                          
                    return false;
                }
            }
            
            if ( index0 == -1 || index1 == -1 )
		return false;
	    if ( index0 != index1 ) {                               
                return true;
            }
	}
    }


    /**
     * @return non-null for FindTaclets
     */
    protected PosInOccurrence getPosInOccurrence ( Goal p_goal ) {
    	final PosInOccurrence topPos =
    	    p_goal.getFormulaTagManager().getPosForTag(positionTag);

	assert topPos != null;
	
	return applicationPosition.replaceConstrainedFormula
	    ( topPos.constrainedFormula () );
    }

}
