// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


package de.uka.ilkd.key.strategy.feature;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.PosInOccurrence;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.rule.TacletApp;
import de.uka.ilkd.key.rule.metaconstruct.arith.Monomial;
import de.uka.ilkd.key.strategy.termProjection.ProjectionToTerm;

/**
 * Feature that decides whether the multiplication of two inequations (using
 * rules of set inEqSimp_nonLin_multiply) is allowed. We only do this if the
 * product of the left sides of the inequations has factors in common with the
 * left side of some other inequation, similarly to how it is decided in the
 * Buchberger algorithm. Also compare with
 * <code>BoundedInEquationsFeature</code>.
 */
public class AllowInEquationMultiplication extends BinaryTacletAppFeature {

    private final ProjectionToTerm targetCandidate;
    private final ProjectionToTerm mult1Candidate;
    private final ProjectionToTerm mult2Candidate;
    
    /**
     * @param mult1Candidate
     *            the left side of the first inequation to be multiplied
     * @param mult2Candidate
     *            the left side of the second inequation to be multiplied
     * @param targetCandidate
     *            the left side of the inequation that is supposed to bound the
     *            other two inequations
     */
    public static Feature create(ProjectionToTerm mult1Candidate,
                                 ProjectionToTerm mult2Candidate,
                                 ProjectionToTerm targetCandidate) {
        return new AllowInEquationMultiplication ( mult1Candidate,
                                                   mult2Candidate,
                                                   targetCandidate );
    }

    private AllowInEquationMultiplication(ProjectionToTerm mult1Candidate,
                                          ProjectionToTerm mult2Candidate,
                                          ProjectionToTerm targetCandidate) {
        this.mult1Candidate = mult1Candidate;
        this.mult2Candidate = mult2Candidate;
        this.targetCandidate = targetCandidate;
    }

    protected boolean filter(TacletApp app, PosInOccurrence pos, Goal goal) {
        final Services services = goal.proof ().getServices ();
        final Monomial targetM =
            Monomial.create ( targetCandidate.toTerm ( app, pos, goal ),
                              services );
        final Monomial mult1M =
            Monomial.create ( mult1Candidate.toTerm ( app, pos, goal ),
                              services );                
        final Monomial mult2M =
            Monomial.create ( mult2Candidate.toTerm ( app, pos, goal ),
                              services );
        
        return !mult2M.divide ( targetM ).variablesDisjoint ( mult1M )
               && !mult1M.divide ( targetM ).variablesDisjoint ( mult2M );
//        return targetM.variablesSubsume ( mult1M.multiply ( mult2M ) );
    }
}
