// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


package de.uka.ilkd.key.strategy.termgenerator;

import de.uka.ilkd.key.logic.IteratorOfTerm;
import de.uka.ilkd.key.logic.ListOfTerm;
import de.uka.ilkd.key.logic.PosInOccurrence;
import de.uka.ilkd.key.logic.SLListOfTerm;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.rule.RuleApp;
import de.uka.ilkd.key.strategy.TopRuleAppCost;
import de.uka.ilkd.key.strategy.termProjection.ProjectionToTerm;
import de.uka.ilkd.key.strategy.termfeature.TermFeature;

/**
 * Term generator that enumerates the subterms or subformulas of a given term.
 * Similarly to <code>RecSubTermFeature</code>, a term feature can be given
 * that determines when traversal should be stopped, i.e., when one should not
 * descend further into a term.
 */
public class SubtermGenerator implements TermGenerator {

    private final TermFeature cond;
    private final ProjectionToTerm completeTerm;

    private SubtermGenerator(ProjectionToTerm completeTerm, TermFeature cond) {
        this.cond = cond;
        this.completeTerm = completeTerm;
    }
    
    public static TermGenerator create(ProjectionToTerm completeTerm,
                                       TermFeature cond) {
        return new SubtermGenerator ( completeTerm, cond );
    }

    public IteratorOfTerm generate(RuleApp app, PosInOccurrence pos, Goal goal) {
        return new Iterator ( completeTerm.toTerm ( app, pos, goal ) );
    }

    private boolean descendFurther(Term t) {
        return ! ( cond.compute ( t ) instanceof TopRuleAppCost );
    }
    
    private class Iterator implements IteratorOfTerm {
        private ListOfTerm termStack;

        public Iterator(Term t) {
            termStack = SLListOfTerm.EMPTY_LIST.prepend ( t );
        }

        public boolean hasNext() {
            return !termStack.isEmpty ();
        }

        public Term next() {
            final Term res = termStack.head ();
            termStack = termStack.tail ();
            
            if ( descendFurther ( res ) ) {
                for ( int i = 0; i != res.arity (); ++i )
                    termStack = termStack.prepend ( res.sub ( i ) );
            }
            
            return res;
        }
    }
}
